import sys

from cpl_core.configuration import ConfigurationABC
from cpl_core.console import Console
from cpl_query.extension import List, IterableABC

from py_to_uxf_core.abc.umlet_creator_abc import UmletCreatorABC
from py_to_uxf_core.configuration.parser_settings import ParserSettings
from py_to_uxf_core.configuration.uml_creator_settings import UMLCreatorSettings
from py_to_uxf_core.model.class_implementation import ClassImplementation
from py_to_uxf_core.model.dimension import Dimension
from py_to_uxf_core.model.position import Position
from py_to_uxf_core.model.python_class import PythonClass
from py_to_uxf_core.model.uml_class import UMLClass

OrderedImplementation = dict[ClassImplementation, IterableABC]


class UmletCreatorService(UmletCreatorABC):

    def __init__(self, config: ConfigurationABC):
        UmletCreatorABC.__init__(self)

        self._settings: UMLCreatorSettings = config.get_configuration(UMLCreatorSettings)
        self._parser_settings: ParserSettings = config.get_configuration(ParserSettings)
        self._moved_classes: list[UMLClass] = []
        self._padding = 30

    @staticmethod
    def _resolve_implementations(classes: List[UMLClass], implementations: List[ClassImplementation]) -> List[UMLClass]:
        new_classes = List(UMLClass)
        sub_classes = List(UMLClass)
        for implementation in implementations:
            implementation: ClassImplementation = implementation
            base: UMLClass = classes.where(lambda c: c.cls == implementation.base).first()
            sub: UMLClass = classes.where(lambda c: c.cls == implementation.subclass).first()

            base.sub_classes.append(sub)
            sub.base_classes.append(base)
            if base not in new_classes:
                new_classes.append(base)
            sub_classes.append(sub)

        for cls in classes:
            if cls not in new_classes and cls not in sub_classes:
                new_classes.append(cls)

        return new_classes

    def _move_classes(self, classes: List[UMLClass], moved_classes: List[UMLClass], next_pos: Position, highest_y: int):
        for cls in classes:
            if cls.cls.name in self._parser_settings.ignore_class_names:
                continue

            if cls not in moved_classes:
                Console.write_line(f'MOVE {cls.cls.name} to {next_pos} - {cls.dimension}')

                if next_pos.x > self._settings.max_pixel_x:
                    next_pos.x = self._padding
                    next_pos.y = highest_y + self._padding * 2

                cls.position.x = next_pos.x
                cls.position.y = next_pos.y

                next_pos.x = cls.position.x + cls.dimension.width + self._padding

                if cls.position.y + cls.dimension.height > highest_y:
                    highest_y = cls.position.y + cls.dimension.height

                moved_classes.append(cls)

            if cls.sub_classes.count() > 0:
                next_pos.x = cls.position.x
                next_pos.y = cls.position.y + cls.dimension.height + self._padding
                self._move_classes(cls.sub_classes, moved_classes, next_pos, highest_y)

    def generate_xml(self, classes: List[PythonClass], implementations: List[ClassImplementation]) -> str:
        uml_classes = List(UMLClass)
        xml_classes = ''
        xml_relations = ''

        for cls in classes:
            uml_cls = UMLClass(cls, Position(0, self._padding), Dimension(80, 80))
            # trigger xml generation to render coordinates
            uml_cls.create_xml_body()
            uml_classes.append(uml_cls)

        sorted_classes = self._resolve_implementations(uml_classes, implementations)
        moved_classes = List(UMLClass)
        self._move_classes(sorted_classes, moved_classes, Position(self._padding, self._padding), 0)

        for uml_cls in moved_classes:
            # save class xml
            if uml_cls.cls.name in self._parser_settings.ignore_class_names:
                continue

            xml_classes += uml_cls.as_xml()

        # self._sort_by_implementation(uml_classes, implementations)
        # oi = self._sort_implementations_by_base_order(implementations)
        # for uml_cls in self._move_by_implementation(oi, uml_classes):
            # save class xml
            # if uml_cls.cls.name in self._parser_settings.ignore_class_names:
            #     continue
            #
            # xml_classes += uml_cls.as_xml()

        # for implementation in implementations:
        #     sub_class: UMLClass = uml_classes.where(lambda u: u.cls == implementation.subclass).first()
        #     base_class: UMLClass = uml_classes.where(lambda u: u.cls == implementation.base).first()
        #     if base_class.cls.name in self._parser_settings.ignore_class_names:
        #         continue
        #     xml_relations += sub_class.implementation_as_xml(base_class)

        return f"""\
        <?xml version="1.0" encoding="UTF-8" standalone="no"?>
        <diagram program="umlet" version="14.1.1">
            <zoom_level>10</zoom_level>
            {xml_classes}
            {xml_relations}
        </diagram>
        """.replace('    ', '').replace('\t', '')
